# -*- coding: utf-8 -*-

import shutil
import tempfile

from pydal import DAL, Field
from pydal._compat import PY2

from ._adapt import DEFAULT_URI, IS_GAE, IS_IMAP, IS_MSSQL, IS_ORACLE, IS_TERADATA, drop
from ._compat import unittest


@unittest.skipIf(IS_IMAP, "Reference not Null unsupported on IMAP")
@unittest.skipIf(IS_ORACLE, "Reference Not Null unsupported on Oracle")
class TestReferenceNOTNULL(unittest.TestCase):
    # 1:N not null

    def testRun(self):
        for ref, bigint in [("reference", False), ("big-reference", True)]:
            db = DAL(DEFAULT_URI, check_reserved=["all"], bigint_id=bigint)
            if bigint and "big-id" not in db._adapter.types:
                continue
            db.define_table("tt", Field("vv"))
            db.define_table(
                "ttt", Field("vv"), Field("tt_id", "%s tt" % ref, notnull=True)
            )
            self.assertRaises(Exception, db.ttt.insert, vv="pydal")
            # The following is mandatory for backends as PG to close the aborted transaction
            db.commit()
            drop(db.ttt)
            drop(db.tt)
            db.close()


@unittest.skipIf(IS_IMAP, "Reference Unique unsupported on IMAP")
@unittest.skipIf(IS_GAE, "Reference Unique unsupported on GAE")
@unittest.skipIf(IS_ORACLE, "Reference Unique unsupported on Oracle")
class TestReferenceUNIQUE(unittest.TestCase):
    # 1:1 relation

    def testRun(self):
        for ref, bigint in [("reference", False), ("big-reference", True)]:
            db = DAL(DEFAULT_URI, check_reserved=["all"], bigint_id=bigint)
            if bigint and "big-id" not in db._adapter.types:
                continue
            db.define_table("tt", Field("vv"))
            db.define_table(
                "ttt",
                Field("vv"),
                Field("tt_id", "%s tt" % ref, unique=True),
                Field("tt_uq", "integer", unique=True),
            )
            id_1 = db.tt.insert(vv="pydal")
            id_2 = db.tt.insert(vv="pydal")
            # Null tt_id
            db.ttt.insert(vv="pydal", tt_uq=1)
            # first insert is OK
            db.ttt.insert(tt_id=id_1, tt_uq=2)
            self.assertRaises(Exception, db.ttt.insert, tt_id=id_1, tt_uq=3)
            self.assertRaises(Exception, db.ttt.insert, tt_id=id_2, tt_uq=2)
            # The following is mandatory for backends as PG to close the aborted transaction
            db.commit()
            drop(db.ttt)
            drop(db.tt)
            db.close()


@unittest.skipIf(IS_IMAP, "Reference Unique not Null unsupported on IMAP")
@unittest.skipIf(IS_GAE, "Reference Unique not Null unsupported on GAE")
@unittest.skipIf(IS_ORACLE, "Reference Unique not Null unsupported on Oracle")
class TestReferenceUNIQUENotNull(unittest.TestCase):
    # 1:1 relation not null

    def testRun(self):
        for ref, bigint in [("reference", False), ("big-reference", True)]:
            db = DAL(DEFAULT_URI, check_reserved=["all"], bigint_id=bigint)
            if bigint and "big-id" not in db._adapter.types:
                continue
            db.define_table("tt", Field("vv"))
            db.define_table(
                "ttt",
                Field("vv"),
                Field("tt_id", "%s tt" % ref, unique=True, notnull=True),
            )
            self.assertRaises(Exception, db.ttt.insert, vv="pydal")
            db.commit()
            id_i = db.tt.insert(vv="pydal")
            # first insert is OK
            db.ttt.insert(tt_id=id_i)
            self.assertRaises(Exception, db.ttt.insert, tt_id=id_i)
            # The following is mandatory for backends as PG to close the aborted transaction
            db.commit()
            drop(db.ttt)
            drop(db.tt)
            db.close()


@unittest.skipIf(IS_IMAP, "Skip unicode on IMAP")
@unittest.skipIf(IS_MSSQL and not PY2, "Skip unicode on py3 and MSSQL")
class TestUnicode(unittest.TestCase):
    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=["all"])
        db.define_table("tt", Field("vv"))
        vv = "ἀγοραζε"
        id_i = db.tt.insert(vv=vv)
        row = db(db.tt.id == id_i).select().first()
        self.assertEqual(row.vv, vv)
        db.commit()
        drop(db.tt)
        db.close()


class TestParseDateTime(unittest.TestCase):
    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=["all"])

        #: skip for adapters that use drivers for datetime parsing
        if db._adapter.parser.registered.get("datetime") is None:
            return

        parse = lambda v: db._adapter.parser.parse(v, "datetime", "datetime")

        dt = parse("2015-09-04t12:33:36.223245")
        self.assertEqual(dt.microsecond, 223245)
        self.assertEqual(dt.hour, 12)

        dt = parse("2015-09-04t12:33:36.223245Z")
        self.assertEqual(dt.microsecond, 223245)
        self.assertEqual(dt.hour, 12)

        dt = parse("2015-09-04t12:33:36.223245-2:0")
        self.assertEqual(dt.microsecond, 223245)
        self.assertEqual(dt.hour, 10)

        dt = parse("2015-09-04t12:33:36+1:0")
        self.assertEqual(dt.microsecond, 0)
        self.assertEqual(dt.hour, 13)

        dt = parse("2015-09-04t12:33:36.123")
        self.assertEqual(dt.microsecond, 123000)

        dt = parse("2015-09-04t12:33:36.00123")
        self.assertEqual(dt.microsecond, 1230)

        dt = parse("2015-09-04t12:33:36.1234567890")
        self.assertEqual(dt.microsecond, 123456)
        db.close()


@unittest.skipIf(IS_IMAP, "chained join unsupported on IMAP")
@unittest.skipIf(IS_TERADATA, "chained join unsupported on TERADATA")
class TestChainedJoinUNIQUE(unittest.TestCase):
    # 1:1 relation

    def testRun(self):
        db = DAL(DEFAULT_URI, check_reserved=["all"])
        db.define_table("aa", Field("name"))
        db.define_table("bb", Field("aa", "reference aa"), Field("name"))
        self.assertEqual(db.bb.aa.referenced_table(), db.aa)
        self.assertEqual(db.bb.aa.referenced_field(), db.aa.id)
        for k in ("x", "y", "z"):
            i = db.aa.insert(name=k)
            for j in ("u", "v", "w"):
                db.bb.insert(aa=i, name=k + j)
        db.commit()
        rows = db(db.aa).select()
        rows.join(db.bb.aa, fields=[db.bb.name], orderby=[db.bb.name])
        self.assertEqual(rows[0].bb[0].name, "xu")
        self.assertEqual(rows[0].bb[1].name, "xv")
        self.assertEqual(rows[0].bb[2].name, "xw")
        self.assertEqual(rows[1].bb[0].name, "yu")
        self.assertEqual(rows[1].bb[1].name, "yv")
        self.assertEqual(rows[1].bb[2].name, "yw")
        self.assertEqual(rows[2].bb[0].name, "zu")
        self.assertEqual(rows[2].bb[1].name, "zv")
        self.assertEqual(rows[2].bb[2].name, "zw")

        rows = db(db.bb).select()
        rows.join(db.aa.id, fields=[db.aa.name])

        self.assertEqual(rows[0].aa.name, "x")
        self.assertEqual(rows[1].aa.name, "x")
        self.assertEqual(rows[2].aa.name, "x")
        self.assertEqual(rows[3].aa.name, "y")
        self.assertEqual(rows[4].aa.name, "y")
        self.assertEqual(rows[5].aa.name, "y")
        self.assertEqual(rows[6].aa.name, "z")
        self.assertEqual(rows[7].aa.name, "z")
        self.assertEqual(rows[8].aa.name, "z")

        rows_json = rows.as_json()
        drop(db.bb)
        drop(db.aa)
        db.close()


class TestNullAdapter(unittest.TestCase):
    # Test that NullAdapter can define tables

    def testRun(self):
        db = DAL(None)
        db.define_table("no_table", Field("aa"))
        self.assertIsInstance(db.no_table.aa, Field)
        self.assertIsInstance(db.no_table["aa"], Field)
        db.close()


@unittest.skip("")
class TestSingleTransaction(unittest.TestCase):
    def test_single_transaction(self):
        db = DAL(DEFAULT_URI)
        db.define_table("tt", Field("aa"))
        self.assertEqual(db(db.tt).count(), 0)
        db.commit()
        try:
            with db.single_transaction():
                db.tt.insert(aa="test")
                1 / 0
        except ZeroDivisionError:
            pass
        self.assertEqual(db(db.tt).count(), 0)
        with db.single_transaction():
            db.tt.insert(aa="test")
        self.assertEqual(db(db.tt).count(), 1)



import os
import re
import time
from urllib.parse import parse_qs

from .._compat import pjoin
from .._globals import THREAD_LOCAL
from ..helpers.classes import SQLALL, FakeDriver, Reference
from ..helpers.methods import use_common_filters, xorify
from ..migrator import InDBMigrator
from ..objects import Expression, Field, Query, Table
from . import adapters, with_connection_or_raise
from .base import NoSQLAdapter
from .mysql import MySQL
from .postgres import PostgrePsyco

try:
    from google.cloud import firestore
except ImportError:
    pass
else:
    from google.cloud.firestore_v1 import aggregation


class GoogleMigratorMixin(object):
    migrator_cls = InDBMigrator


class Streamable:
    def __init__(self, item):
        self.item = item

    def stream(self):
        if self.item:
            yield self.item


@adapters.register_for("google:sql")
class GoogleSQL(GoogleMigratorMixin, MySQL):
    uploads_in_blob = True
    REGEX_URI = "^(?P<instance>.*)/(?P<db>.+)$"

    def _find_work_folder(self):
        super(GoogleSQL, self)._find_work_folder()
        if os.path.isabs(self.folder) and self.folder.startswith(os.getcwd()):
            self.folder = os.path.relpath(self.folder, os.getcwd())

    def _initialize_(self):
        super(GoogleSQL, self)._initialize_()
        self.folder = self.folder or pjoin(
            "$HOME",
            THREAD_LOCAL._pydal_folder_.split(os.sep + "applications" + os.sep, 1)[1],
        )
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        self.driver_args["instance"] = self.credential_decoder(m.group("instance"))
        self.dbstring = self.credential_decoder(m.group("db"))
        self.createdb = self.adapter_args.get("createdb", True)
        if not self.createdb:
            self.driver_args["database"] = self.dbstring

    def find_driver(self):
        self.driver = "google"

    def connector(self):
        return rdbms.connect(**self.driver_args)

    def after_connection(self):
        if self.createdb:
            self.execute("CREATE DATABASE IF NOT EXISTS %s" % self.dbstring)
            self.execute("USE %s" % self.dbstring)
        self.execute("SET FOREIGN_KEY_CHECKS=1;")
        self.execute("SET sql_mode='NO_BACKSLASH_ESCAPES';")

    @with_connection_or_raise
    def execute(self, *args, **kwargs):
        command = self.filter_sql_command(args[0]).decode("utf8")
        handlers = self._build_handlers_for_execution()
        for handler in handlers:
            handler.before_execute(command)
        rv = self.cursor.execute(command, *args[1:], **kwargs)
        for handler in handlers:
            handler.after_execute(command)
        return rv


# based on this: https://cloud.google.com/appengine/docs/standard/python/cloud-sql/
@adapters.register_for("google:MySQLdb")
class GoogleMySQL(GoogleMigratorMixin, MySQL):
    uploads_in_blob = True
    drivers = ("MySQLdb",)

    def _find_work_folder(self):
        super(GoogleMySQL, self)._find_work_folder()
        if os.path.isabs(self.folder) and self.folder.startswith(os.getcwd()):
            self.folder = os.path.relpath(self.folder, os.getcwd())

    def after_connection(self):
        self.execute("SET FOREIGN_KEY_CHECKS=1;")
        self.execute("SET sql_mode='NO_BACKSLASH_ESCAPES,TRADITIONAL';")


@adapters.register_for("google:psycopg2")
class GooglePostgres(GoogleMigratorMixin, PostgrePsyco):
    uploads_in_blob = True
    drivers = ("psycopg2",)

    def _find_work_folder(self):
        super(GooglePostgres, self)._find_work_folder()
        if os.path.isabs(self.folder) and self.folder.startswith(os.getcwd()):
            self.folder = os.path.relpath(self.folder, os.getcwd())


# db = DAL("firestore") on Google App Engine
# db = DAL("firestore://cred=credentials.json") everywhere else
# (get credentials.json from google console)
#
# supports:
# db(db.table).select(...)
# db(db.table.id==id).select(...)
# db(db.table.id>0).select(...)
# db(db.table.field==value).select(...)
# db(db.table.field!=value).select(...)
# db(db.table.field>value).select(...)
# db(db.table.field<value).select(...)
# db(db.table.field>=value).select(...)
# db(db.table.field<=value).select(...)
# db(db.table.field.belnogs(values)).select(...)
# db(~db.table.field.belnogs(values)).select(...)
# db(simple_query1|simple_query2|simple_query3).select(...)
# db(simple_query1&simple_query2&simple_query3).select(...)
# db(or_query1&or_query2&or_query3).select(...)
# db.table(id)
# db.table(field1=value, field2=value)
# db.table(id, field1=value, field2=value)
# db(...).delete()
# db(...).update(...) but only reliable when updating one or a few records
#
# does not support:
# - joins
# - transaction
# - composition of queries involving field id
# - or of and such as db(query1|(query2&query2))
# - db.table.field.contains(...)
# - db.table.field.like(...)
# - db.table.field.startswith(...)
# - non-zero offset such as select(limitby=(3,7))
#
# supports but unreliable
# db(...).update(...) for many records


@adapters.register_for("firestore")
class Firestore(NoSQLAdapter):
    dbengine = "firestore"

    def _initialize_(self):
        super(Firestore, self)._initialize_()
        args = parse_qs(self.uri.split("//")[1]) if "//" in self.uri else {}
        self._client = firestore.Client(project=args.get("id", [None])[0])

    def find_driver(self):
        return

    def connector(self):
        return FakeDriver()

    def _expand(self, expression, field_type=None, query_env={}):
        if expression is None:
            return None
        elif isinstance(expression, Field):
            if expression.type in ("text", "blob", "json"):
                raise SyntaxError("AppEngine does not index by: %s" % expression.type)
            return expression.name
        elif isinstance(expression, (Expression, Query)):
            if expression.second is not None:
                return expression.op(
                    expression.first, expression.second, query_env=query_env
                )
            elif expression.first is not None:
                return expression.op(expression.first, query_env=query_env)
            else:
                return expression.op()
        elif field_type:
            return self.represent(expression, field_type)
        elif isinstance(expression, (list, tuple)):
            return ",".join([self.represent(item, field_type) for item in expression])
        elif hasattr(expression, "_FilterNode__name"):
            # check for _FilterNode__name to avoid explicit
            # import of FilterNode
            return expression
        else:
            raise NotImplementedError

    def represent(self, obj, field_type, tablename=None):
        if isinstance(obj, (Expression, Field)):
            raise SyntaxError("not supported on GAE")
        return super(Firestore, self).represent(obj, field_type)

    def apply_filter(self, source, table, query):
        if isinstance(query, Query) and query.first is table._id:
            if query.op.__name__ == "eq":
                return source.document(str(query.second)).get()
            elif isinstance(query, Table):
                return source
            elif query.op.__name__ == "gt" and query.second == 0:
                return source
            raise RuntimeError("operator not supported")

        filters = self.expand(query)
        if filters:
            if not isinstance(filters, list):
                filters = [filters]
            for filter in filters:
                source = source.where(filter=filter)
        return source

    def get_docs(self, table, query, orderby=None, limitby=None):
        source = self._client.collection(table._tablename)
        source = self.apply_filter(source, table, query)

        if not hasattr(source, "where"):
            return [source]

        if limitby:
            if limitby[0]:
                raise SyntaxError("Firestore does not support an offset")
            source = source.limit(limitby[1])

        if orderby:
            if not isinstance(orderby, list):
                orderby = [orderby]
            for order in orderby:
                if isinstance(order, Field):
                    source = source.order_by(order.name)
                elif (
                    isinstance(order, Expression)
                    and order.op.__name__ == "invert"
                    and isinstance(order.first, Field)
                ):
                    source = source.order_by(
                        order.first.name, direction=firestore.Query.DESCENDING
                    )
                else:
                    raise RuntimeError(f"orderby {order} unsupported")

        return source.stream()

    def select(self, query, fields, attributes):
        attributes = attributes or {}

        # figure out what query and fields we want
        if query:
            table = self.get_table(query)
        elif fields:
            table = fields[0].table
            query = db._adapter.id_query(fields[0].table)
        else:
            raise SyntaxError("Unable to determine the table")

        if fields and not isinstance(fields, (list, tuple)):
            fields = [fields]
        if not fields or SQLALL in fields:
            fields = list(table)

        orderby = attributes.get("orderby")
        limitby = attributes.get("limitby")
        if attributes.get("distinct"):
            raise SyntaxError("Firestore does not support an distinct")
        docs = self.get_docs(table, query, orderby, limitby)

        # convert docs to rows
        rows = []
        for doc in docs:
            item = doc.to_dict()
            if not item:
                continue
            item["id"] = int(doc.id)
            rows.append([item.get(t.name) for t in fields])
        # postprocess the items
        colnames = [t.longname for t in fields]
        processor = attributes.get("processor", self.parse)
        return processor(rows, fields, colnames, False)

    def count(self, query, distinct=None, limit=None):
        # OK
        if distinct:
            raise RuntimeError("COUNT DISTINCT not supported")
        table = self.get_table(query)
        source = self._client.collection(table._tablename)
        source = self.apply_filter(source, table, query)
        # deal with the case or returning a single item which may be empty
        if not hasattr(source, "where"):
            return 1 if source.to_dict() else 0
        aggregate_query = aggregation.AggregationQuery(source)
        aggregate_query.count()
        results = aggregate_query.get()
        for result in results:
            return int(result[0].value)
        return 0

    def delete(self, table, query):
        while self.db(query).count() > 0:
            docs = list(self.get_docs(table, query))
            batch = self._client.batch()
            counter = 0
            for doc in docs:
                batch.delete(doc.reference)
                counter += 1
            if counter:
                batch.commit()
        return counter

    def update(self, table, query, update_fields):
        counter = 0
        if any(f.name == "id" for f, v in update_fields):
            raise RuntimeError("Cannot update the id field")
        docs = list(self.get_docs(table, query))
        batch = self._client.batch()
        for doc in docs:
            batch.update(doc.reference, {f.name: v for f, v in update_fields})
            counter += 1
        try:
            batch.commit()
        except Exception:
            pass  # google.api_core.exceptions.NotFound: No document to update
        return counter

    def truncate(self, table, mode=""):
        def delete_collection(coll_ref, batch_size):
            if batch_size == 0:
                return
            batch = self._client.batch()
            docs = coll_ref.list_documents(page_size=batch_size)
            deleted = 0
            for doc in docs:
                batch.delete(doc)
                deleted = deleted + 1
            if deleted:
                batch.commit()
            if deleted >= batch_size:
                return delete_collection(coll_ref, batch_size)

        delete_collection(self._client.collection(table._tablename), 1000)

    @staticmethod
    def make_id():
        return int(time.time() * int(1e21) + int.from_bytes(os.urandom(8)) % int(1e18))

    def insert(self, table, fields):
        # OK
        if any(f.name == "id" for f, v in fields):
            raise RuntimeError("Cannot update the id field")
        dfields = dict((f.name, self.represent(v, f.type)) for f, v in fields)
        id = self.make_id()
        collection = self._client.collection(table._tablename)
        doc = collection.document(str(id))
        doc.set(dfields)
        rid = Reference(id)
        rid._table, rid._record = table, None
        return rid

    def bulk_insert(self, table, items):
        # OK
        collection = self._client.collection(table._tablename)
        batch = self._client.batch()
        ids = []
        for fields in items:
            if any(f.name == "id" for f, v in fields):
                raise RuntimeError("Cannot update the id field")
            dfields = dict((f.name, self.represent(v, f.type)) for f, v in fields)
            id = self.make_id()
            doc = collection.document(str(id))
            batch.set(doc, dfields)
            rid = Reference(id)
            rid._table, rid._recor = table, None
            ids.append(rid)
        if items:
            batch.commit()
        return ids



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Unit tests for scheduler.py"""

import datetime
import tempfile
import time
import unittest

from pydal import DAL
from pydal.tools.scheduler import Scheduler, delta, now


class TestScheduler(unittest.TestCase):
    def test_scheduler(self):
        completed_tasks = []

        def mytask():
            print(f"hello world {datetime.datetime.now()}")
            return "Done!"

        with tempfile.TemporaryDirectory() as tempdir:
            db = DAL("sqlite://store.sqllite", folder=tempdir)
            scheduler = Scheduler(db, sleep_time=1, max_concurrent_runs=1)
            db(db.task_run).delete()
            # a task that says hello
            scheduler.register_task("hello", mytask)
            scheduler.register_task("periodic", mytask)
            # a task that fails if x is 0
            scheduler.register_task("fail", lambda x: 1 / x)
            # task that takes 10 seconds
            scheduler.register_task("long", lambda: time.sleep(10))
            # enqueue a run for the fail task with input x=0 (for indeed fails)
            scheduler.enqueue_run(name="fail", inputs={"x": 0})
            # a run that takes 10s but with timeout of 1s so timeouts
            scheduler.enqueue_run(name="long", timeout=1)
            # a run scheduled to start in 20s
            scheduler.enqueue_run(name="hello", scheduled_for=now() + delta(2))
            # a regular run to start asap
            scheduler.enqueue_run(name="hello")
            # a run that should rerun every 10s
            scheduler.enqueue_run(name="periodic", period=2)
            db.commit()
            scheduler.start()
            while True:
                # do not all more than 3 periodic tasks
                if db(db.task_run.name == "periodic").count() > 3:
                    db(db.task_run.name == "periodic")(
                        db.task_run.status == "queued"
                    ).delete()
                    db.commit()
                    break
                time.sleep(0.1)
            scheduler.stop()
            self.assertEqual(db(db.task_run).count(), 7)
            self.assertEqual(db(db.task_run.status == "queued").count(), 0)
            self.assertEqual(db(db.task_run.status == "timeout").count(), 1)
            self.assertEqual(db(db.task_run.status == "completed").count(), 5)
            self.assertEqual(db(db.task_run.status == "failed").count(), 1)
            # for run in db(db.task_run).select():
            #     print(run.name, run.status, run.log)



import re

from ..helpers._internals import Dispatcher


class Adapters(Dispatcher):
    def register_for(self, *uris):
        def wrap(dispatch_class):
            for uri in uris:
                self._registry_[uri] = dispatch_class
            return dispatch_class

        return wrap

    def get_for(self, uri):
        try:
            return self._registry_[uri]
        except KeyError:
            raise SyntaxError("Adapter not found for %s" % uri)


adapters = Adapters("adapters")


class AdapterMeta(type):
    """Metaclass to support manipulation of adapter classes.

    At the moment is used to intercept `entity_quoting` argument passed to DAL.
    """

    def __call__(cls, *args, **kwargs):
        uploads_in_blob = kwargs.get("adapter_args", {}).get(
            "uploads_in_blob", cls.uploads_in_blob
        )
        cls.uploads_in_blob = uploads_in_blob

        entity_quoting = kwargs.get("entity_quoting", True)
        if "entity_quoting" in kwargs:
            del kwargs["entity_quoting"]

        obj = super(AdapterMeta, cls).__call__(*args, **kwargs)

        regex_ent = r"(\w+)"
        if not entity_quoting:
            obj.dialect.quote_template = "%s"
        else:
            regex_ent = obj.dialect.quote_template % regex_ent
        # FIXME: this regex should NOT be compiled
        obj.REGEX_TABLE_DOT_FIELD = re.compile(r"^%s\.%s$" % (regex_ent, regex_ent))

        return obj


def with_connection(f):
    def wrap(*args, **kwargs):
        if args[0].connection:
            return f(*args, **kwargs)
        return None

    return wrap


def with_connection_or_raise(f):
    def wrap(*args, **kwargs):
        if not args[0].connection:
            if len(args) > 1:
                raise ValueError(args[1])
            raise RuntimeError("no connection available")
        return f(*args, **kwargs)

    return wrap


from .base import NoSQLAdapter, SQLAdapter
from .couchdb import CouchDB
from .db2 import DB2
from .firebird import FireBird
from .google import GoogleSQL
from .informix import Informix
from .ingres import Ingres
from .mongo import Mongo
from .mssql import MSSQL
from .mysql import MySQL
from .oracle import Oracle
from .postgres import Postgre, PostgrePsyco
from .sap import SAPDB
from .snowflake import Snowflake
from .sqlite import SQLite
from .teradata import Teradata



# encoding utf-8

__author__ = "Thadeus Burgess <thadeusb@thadeusb.com>"

#    we classify as "non-reserved" those key words that are explicitly known
#    to the parser but are allowed as column or table names. Some key words
#    that are otherwise non-reserved cannot be used as function or data type n
#    ames and are in the nonreserved list. (Most of these words represent
#    built-in functions or data types with special syntax. The function
#    or type is still available but it cannot be redefined by the user.)
#    Labeled "reserved" are those tokens that are not allowed as column or
#    table names. Some reserved key words are allowable as names for
#    functions or data typesself.

# Note at the bottom of the list is a dict containing references to the
# tuples, and also if you add a list don't forget to remove its default
# set of COMMON.

# Keywords that are adapter specific. Such as a list of "postgresql"
# or "mysql" keywords

# These are keywords that are common to all SQL dialects, and should
# never be used as a table or column. Even if you use one of these
# the cursor will throw an OperationalError for the SQL syntax.

from .._compat import PY2

if not PY2:
    from functools import reduce

COMMON = set(
    (
        "SELECT",
        "INSERT",
        "DELETE",
        "UPDATE",
        "DROP",
        "CREATE",
        "ALTER",
        "WHERE",
        "FROM",
        "INNER",
        "JOIN",
        "AND",
        "OR",
        "LIKE",
        "ON",
        "IN",
        "SET",
        "BY",
        "GROUP",
        "ORDER",
        "LEFT",
        "OUTER",
        "IF",
        "END",
        "THEN",
        "LOOP",
        "AS",
        "ELSE",
        "FOR",
        "CASE",
        "WHEN",
        "MIN",
        "MAX",
        "DISTINCT",
    )
)


POSTGRESQL = set(
    (
        "FALSE",
        "TRUE",
        "ALL",
        "ANALYSE",
        "ANALYZE",
        "AND",
        "ANY",
        "ARRAY",
        "AS",
        "ASC",
        "ASYMMETRIC",
        "AUTHORIZATION",
        "BETWEEN",
        "BIGINT",
        "BINARY",
        "BIT",
        "BOOLEAN",
        "BOTH",
        "CASE",
        "CAST",
        "CHAR",
        "CHARACTER",
        "CHECK",
        "COALESCE",
        "COLLATE",
        "COLUMN",
        "CONSTRAINT",
        "CREATE",
        "CROSS",
        "CURRENT_CATALOG",
        "CURRENT_DATE",
        "CURRENT_ROLE",
        "CURRENT_SCHEMA",
        "CURRENT_TIME",
        "CURRENT_TIMESTAMP",
        "CURRENT_USER",
        "DEC",
        "DECIMAL",
        "DEFAULT",
        "DEFERRABLE",
        "DESC",
        "DISTINCT",
        "DO",
        "ELSE",
        "END",
        "EXCEPT",
        "EXISTS",
        "EXTRACT",
        "FETCH",
        "FLOAT",
        "FOR",
        "FOREIGN",
        "FREEZE",
        "FROM",
        "FULL",
        "GRANT",
        "GREATEST",
        "GROUP",
        "HAVING",
        "ILIKE",
        "IN",
        "INITIALLY",
        "INNER",
        "INOUT",
        "INT",
        "INTEGER",
        "INTERSECT",
        "INTERVAL",
        "INTO",
        "IS",
        "ISNULL",
        "JOIN",
        "LEADING",
        "LEAST",
        "LEFT",
        "LIKE",
        "LIMIT",
        "LOCALTIME",
        "LOCALTIMESTAMP",
        "NATIONAL",
        "NATURAL",
        "NCHAR",
        "NEW",
        "NONE",
        "NOT",
        "NOTNULL",
        "NULL",
        "NULLIF",
        "NUMERIC",
        "OFF",
        "OFFSET",
        "OLD",
        "ON",
        "ONLY",
        "OR",
        "ORDER",
        "OUT",
        "OUTER",
        "OVERLAPS",
        "OVERLAY",
        "PLACING",
        "POSITION",
        "PRECISION",
        "PRIMARY",
        "REAL",
        "REFERENCES",
        "RETURNING",
        "RIGHT",
        "ROW",
        "SELECT",
        "SESSION_USER",
        "SETOF",
        "SIMILAR",
        "SMALLINT",
        "SOME",
        "SUBSTRING",
        "SYMMETRIC",
        "TABLE",
        "THEN",
        "TIME",
        "TIMESTAMP",
        "TO",
        "TRAILING",
        "TREAT",
        "TRIM",
        "UNION",
        "UNIQUE",
        "USER",
        "USING",
        "VALUES",
        "VARCHAR",
        "VARIADIC",
        "VERBOSE",
        "WHEN",
        "WHERE",
        "WITH",
        "XMLATTRIBUTES",
        "XMLCONCAT",
        "XMLELEMENT",
        "XMLFOREST",
        "XMLPARSE",
        "XMLPI",
        "XMLROOT",
        "XMLSERIALIZE",
    )
)


POSTGRESQL_NONRESERVED = set(
    (
        "A",
        "ABORT",
        "ABS",
        "ABSENT",
        "ABSOLUTE",
        "ACCESS",
        "ACCORDING",
        "ACTION",
        "ADA",
        "ADD",
        "ADMIN",
        "AFTER",
        "AGGREGATE",
        "ALIAS",
        "ALLOCATE",
        "ALSO",
        "ALTER",
        "ALWAYS",
        "ARE",
        "ARRAY_AGG",
        "ASENSITIVE",
        "ASSERTION",
        "ASSIGNMENT",
        "AT",
        "ATOMIC",
        "ATTRIBUTE",
        "ATTRIBUTES",
        "AVG",
        "BACKWARD",
        "BASE64",
        "BEFORE",
        "BEGIN",
        "BERNOULLI",
        "BIT_LENGTH",
        "BITVAR",
        "BLOB",
        "BOM",
        "BREADTH",
        "BY",
        "C",
        "CACHE",
        "CALL",
        "CALLED",
        "CARDINALITY",
        "CASCADE",
        "CASCADED",
        "CATALOG",
        "CATALOG_NAME",
        "CEIL",
        "CEILING",
        "CHAIN",
        "CHAR_LENGTH",
        "CHARACTER_LENGTH",
        "CHARACTER_SET_CATALOG",
        "CHARACTER_SET_NAME",
        "CHARACTER_SET_SCHEMA",
        "CHARACTERISTICS",
        "CHARACTERS",
        "CHECKED",
        "CHECKPOINT",
        "CLASS",
        "CLASS_ORIGIN",
        "CLOB",
        "CLOSE",
        "CLUSTER",
        "COBOL",
        "COLLATION",
        "COLLATION_CATALOG",
        "COLLATION_NAME",
        "COLLATION_SCHEMA",
        "COLLECT",
        "COLUMN_NAME",
        "COLUMNS",
        "COMMAND_FUNCTION",
        "COMMAND_FUNCTION_CODE",
        "COMMENT",
        "COMMIT",
        "COMMITTED",
        "COMPLETION",
        "CONCURRENTLY",
        "CONDITION",
        "CONDITION_NUMBER",
        "CONFIGURATION",
        "CONNECT",
        "CONNECTION",
        "CONNECTION_NAME",
        "CONSTRAINT_CATALOG",
        "CONSTRAINT_NAME",
        "CONSTRAINT_SCHEMA",
        "CONSTRAINTS",
        "CONSTRUCTOR",
        "CONTAINS",
        "CONTENT",
        "CONTINUE",
        "CONVERSION",
        "CONVERT",
        "COPY",
        "CORR",
        "CORRESPONDING",
        "COST",
        "COUNT",
        "COVAR_POP",
        "COVAR_SAMP",
        "CREATEDB",
        "CREATEROLE",
        "CREATEUSER",
        "CSV",
        "CUBE",
        "CUME_DIST",
        "CURRENT",
        "CURRENT_DEFAULT_TRANSFORM_GROUP",
        "CURRENT_PATH",
        "CURRENT_TRANSFORM_GROUP_FOR_TYPE",
        "CURSOR",
        "CURSOR_NAME",
        "CYCLE",
        "DATA",
        "DATABASE",
        "DATE",
        "DATETIME_INTERVAL_CODE",
        "DATETIME_INTERVAL_PRECISION",
        "DAY",
        "DEALLOCATE",
        "DECLARE",
        "DEFAULTS",
        "DEFERRED",
        "DEFINED",
        "DEFINER",
        "DEGREE",
        "DELETE",
        "DELIMITER",
        "DELIMITERS",
        "DENSE_RANK",
        "DEPTH",
        "DEREF",
        "DERIVED",
        "DESCRIBE",
        "DESCRIPTOR",
        "DESTROY",
        "DESTRUCTOR",
        "DETERMINISTIC",
        "DIAGNOSTICS",
        "DICTIONARY",
        "DISABLE",
        "DISCARD",
        "DISCONNECT",
        "DISPATCH",
        "DOCUMENT",
        "DOMAIN",
        "DOUBLE",
        "DROP",
        "DYNAMIC",
        "DYNAMIC_FUNCTION",
        "DYNAMIC_FUNCTION_CODE",
        "EACH",
        "ELEMENT",
        "EMPTY",
        "ENABLE",
        "ENCODING",
        "ENCRYPTED",
        "END-EXEC",
        "ENUM",
        "EQUALS",
        "ESCAPE",
        "EVERY",
        "EXCEPTION",
        "EXCLUDE",
        "EXCLUDING",
        "EXCLUSIVE",
        "EXEC",
        "EXECUTE",
        "EXISTING",
        "EXP",
        "EXPLAIN",
        "EXTERNAL",
        "FAMILY",
        "FILTER",
        "FINAL",
        "FIRST",
        "FIRST_VALUE",
        "FLAG",
        "FLOOR",
        "FOLLOWING",
        "FORCE",
        "FORTRAN",
        "FORWARD",
        "FOUND",
        "FREE",
        "FUNCTION",
        "FUSION",
        "G",
        "GENERAL",
        "GENERATED",
        "GET",
        "GLOBAL",
        "GO",
        "GOTO",
        "GRANTED",
        "GROUPING",
        "HANDLER",
        "HEADER",
        "HEX",
        "HIERARCHY",
        "HOLD",
        "HOST",
        "HOUR",
        #    'ID',
        "IDENTITY",
        "IF",
        "IGNORE",
        "IMMEDIATE",
        "IMMUTABLE",
        "IMPLEMENTATION",
        "IMPLICIT",
        "INCLUDING",
        "INCREMENT",
        "INDENT",
        "INDEX",
        "INDEXES",
        "INDICATOR",
        "INFIX",
        "INHERIT",
        "INHERITS",
        "INITIALIZE",
        "INPUT",
        "INSENSITIVE",
        "INSERT",
        "INSTANCE",
        "INSTANTIABLE",
        "INSTEAD",
        "INTERSECTION",
        "INVOKER",
        "ISOLATION",
        "ITERATE",
        "K",
        "KEY",
        "KEY_MEMBER",
        "KEY_TYPE",
        "LAG",
        "LANCOMPILER",
        "LANGUAGE",
        "LARGE",
        "LAST",
        "LAST_VALUE",
        "LATERAL",
        "LC_COLLATE",
        "LC_CTYPE",
        "LEAD",
        "LENGTH",
        "LESS",
        "LEVEL",
        "LIKE_REGEX",
        "LISTEN",
        "LN",
        "LOAD",
        "LOCAL",
        "LOCATION",
        "LOCATOR",
        "LOCK",
        "LOGIN",
        "LOWER",
        "M",
        "MAP",
        "MAPPING",
        "MATCH",
        "MATCHED",
        "MAX",
        "MAX_CARDINALITY",
        "MAXVALUE",
        "MEMBER",
        "MERGE",
        "MESSAGE_LENGTH",
        "MESSAGE_OCTET_LENGTH",
        "MESSAGE_TEXT",
        "METHOD",
        "MIN",
        "MINUTE",
        "MINVALUE",
        "MOD",
        "MODE",
        "MODIFIES",
        "MODIFY",
        "MODULE",
        "MONTH",
        "MORE",
        "MOVE",
        "MULTISET",
        "MUMPS",
        #    'NAME',
        "NAMES",
        "NAMESPACE",
        "NCLOB",
        "NESTING",
        "NEXT",
        "NFC",
        "NFD",
        "NFKC",
        "NFKD",
        "NIL",
        "NO",
        "NOCREATEDB",
        "NOCREATEROLE",
        "NOCREATEUSER",
        "NOINHERIT",
        "NOLOGIN",
        "NORMALIZE",
        "NORMALIZED",
        "NOSUPERUSER",
        "NOTHING",
        "NOTIFY",
        "NOWAIT",
        "NTH_VALUE",
        "NTILE",
        "NULLABLE",
        "NULLS",
        "NUMBER",
        "OBJECT",
        "OCCURRENCES_REGEX",
        "OCTET_LENGTH",
        "OCTETS",
        "OF",
        "OIDS",
        "OPEN",
        "OPERATION",
        "OPERATOR",
        "OPTION",
        "OPTIONS",
        "ORDERING",
        "ORDINALITY",
        "OTHERS",
        "OUTPUT",
        "OVER",
        "OVERRIDING",
        "OWNED",
        "OWNER",
        "P",
        "PAD",
        "PARAMETER",
        "PARAMETER_MODE",
        "PARAMETER_NAME",
        "PARAMETER_ORDINAL_POSITION",
        "PARAMETER_SPECIFIC_CATALOG",
        "PARAMETER_SPECIFIC_NAME",
        "PARAMETER_SPECIFIC_SCHEMA",
        "PARAMETERS",
        "PARSER",
        "PARTIAL",
        "PARTITION",
        "PASCAL",
        "PASSING",
        #    'PASSWORD',
        "PATH",
        "PERCENT_RANK",
        "PERCENTILE_CONT",
        "PERCENTILE_DISC",
        "PLANS",
        "PLI",
        "POSITION_REGEX",
        "POSTFIX",
        "POWER",
        "PRECEDING",
        "PREFIX",
        "PREORDER",
        "PREPARE",
        "PREPARED",
        "PRESERVE",
        "PRIOR",
        "PRIVILEGES",
        "PROCEDURAL",
        "PROCEDURE",
        "PUBLIC",
        "QUOTE",
        "RANGE",
        "RANK",
        "READ",
        "READS",
        "REASSIGN",
        "RECHECK",
        "RECURSIVE",
        "REF",
        "REFERENCING",
        "REGR_AVGX",
        "REGR_AVGY",
        "REGR_COUNT",
        "REGR_INTERCEPT",
        "REGR_R2",
        "REGR_SLOPE",
        "REGR_SXX",
        "REGR_SXY",
        "REGR_SYY",
        "REINDEX",
        "RELATIVE",
        "RELEASE",
        "RENAME",
        "REPEATABLE",
        "REPLACE",
        "REPLICA",
        "RESET",
        "RESPECT",
        "RESTART",
        "RESTRICT",
        "RESULT",
        "RETURN",
        "RETURNED_CARDINALITY",
        "RETURNED_LENGTH",
        "RETURNED_OCTET_LENGTH",
        "RETURNED_SQLSTATE",
        "RETURNS",
        "REVOKE",
        #    'ROLE',
        "ROLLBACK",
        "ROLLUP",
        "ROUTINE",
        "ROUTINE_CATALOG",
        "ROUTINE_NAME",
        "ROUTINE_SCHEMA",
        "ROW_COUNT",
        "ROW_NUMBER",
        "ROWS",
        "RULE",
        "SAVEPOINT",
        "SCALE",
        "SCHEMA",
        "SCHEMA_NAME",
        "SCOPE",
        "SCOPE_CATALOG",
        "SCOPE_NAME",
        "SCOPE_SCHEMA",
        "SCROLL",
        "SEARCH",
        "SECOND",
        "SECTION",
        "SECURITY",
        "SELF",
        "SENSITIVE",
        "SEQUENCE",
        "SERIALIZABLE",
        "SERVER",
        "SERVER_NAME",
        "SESSION",
        "SET",
        "SETS",
        "SHARE",
        "SHOW",
        "SIMPLE",
        "SIZE",
        "SOURCE",
        "SPACE",
        "SPECIFIC",
        "SPECIFIC_NAME",
        "SPECIFICTYPE",
        "SQL",
        "SQLCODE",
        "SQLERROR",
        "SQLEXCEPTION",
        "SQLSTATE",
        "SQLWARNING",
        "SQRT",
        "STABLE",
        "STANDALONE",
        "START",
        "STATE",
        "STATEMENT",
        "STATIC",
        "STATISTICS",
        "STDDEV_POP",
        "STDDEV_SAMP",
        "STDIN",
        "STDOUT",
        "STORAGE",
        "STRICT",
        "STRIP",
        "STRUCTURE",
        "STYLE",
        "SUBCLASS_ORIGIN",
        "SUBLIST",
        "SUBMULTISET",
        "SUBSTRING_REGEX",
        "SUM",
        "SUPERUSER",
        "SYSID",
        "SYSTEM",
        "SYSTEM_USER",
        "T",
        #    'TABLE_NAME',
        "TABLESAMPLE",
        "TABLESPACE",
        "TEMP",
        "TEMPLATE",
        "TEMPORARY",
        "TERMINATE",
        "TEXT",
        "THAN",
        "TIES",
        "TIMEZONE_HOUR",
        "TIMEZONE_MINUTE",
        "TOP_LEVEL_COUNT",
        "TRANSACTION",
        "TRANSACTION_ACTIVE",
        "TRANSACTIONS_COMMITTED",
        "TRANSACTIONS_ROLLED_BACK",
        "TRANSFORM",
        "TRANSFORMS",
        "TRANSLATE",
        "TRANSLATE_REGEX",
        "TRANSLATION",
        "TRIGGER",
        "TRIGGER_CATALOG",
        "TRIGGER_NAME",
        "TRIGGER_SCHEMA",
        "TRIM_ARRAY",
        "TRUNCATE",
        "TRUSTED",
        "TYPE",
        "UESCAPE",
        "UNBOUNDED",
        "UNCOMMITTED",
        "UNDER",
        "UNENCRYPTED",
        "UNKNOWN",
        "UNLISTEN",
        "UNNAMED",
        "UNNEST",
        "UNTIL",
        "UNTYPED",
        "UPDATE",
        "UPPER",
        "URI",
        "USAGE",
        "USER_DEFINED_TYPE_CATALOG",
        "USER_DEFINED_TYPE_CODE",
        "USER_DEFINED_TYPE_NAME",
        "USER_DEFINED_TYPE_SCHEMA",
        "VACUUM",
        "VALID",
        "VALIDATOR",
        "VALUE",
        "VAR_POP",
        "VAR_SAMP",
        "VARBINARY",
        "VARIABLE",
        "VARYING",
        "VERSION",
        "VIEW",
        "VOLATILE",
        "WHENEVER",
        "WHITESPACE",
        "WIDTH_BUCKET",
        "WINDOW",
        "WITHIN",
        "WITHOUT",
        "WORK",
        "WRAPPER",
        "WRITE",
        "XML",
        "XMLAGG",
        "XMLBINARY",
        "XMLCAST",
        "XMLCOMMENT",
        "XMLDECLARATION",
        "XMLDOCUMENT",
        "XMLEXISTS",
        "XMLITERATE",
        "XMLNAMESPACES",
        "XMLQUERY",
        "XMLSCHEMA",
        "XMLTABLE",
        "XMLTEXT",
        "XMLVALIDATE",
        "YEAR",
        "YES",
        "ZONE",
    )
)

# Thanks villas
FIREBIRD = set(
    (
        "ABS",
        "ACTIVE",
        "ADMIN",
        "AFTER",
        "ASCENDING",
        "AUTO",
        "AUTODDL",
        "BASED",
        "BASENAME",
        "BASE_NAME",
        "BEFORE",
        "BIT_LENGTH",
        "BLOB",
        "BLOBEDIT",
        "BOOLEAN",
        "BOTH",
        "BUFFER",
        "CACHE",
        "CHAR_LENGTH",
        "CHARACTER_LENGTH",
        "CHECK_POINT_LEN",
        "CHECK_POINT_LENGTH",
        "CLOSE",
        "COMMITTED",
        "COMPILETIME",
        "COMPUTED",
        "CONDITIONAL",
        "CONNECT",
        "CONTAINING",
        "CROSS",
        "CSTRING",
        "CURRENT_CONNECTION",
        "CURRENT_ROLE",
        "CURRENT_TRANSACTION",
        "CURRENT_USER",
        "DATABASE",
        "DB_KEY",
        "DEBUG",
        "DESCENDING",
        "DISCONNECT",
        "DISPLAY",
        "DO",
        "ECHO",
        "EDIT",
        "ENTRY_POINT",
        "EVENT",
        "EXIT",
        "EXTERN",
        "FALSE",
        "FETCH",
        "FILE",
        "FILTER",
        "FREE_IT",
        "FUNCTION",
        "GDSCODE",
        "GENERATOR",
        "GEN_ID",
        "GLOBAL",
        "GROUP_COMMIT_WAIT",
        "GROUP_COMMIT_WAIT_TIME",
        "HELP",
        "IF",
        "INACTIVE",
        "INDEX",
        "INIT",
        "INPUT_TYPE",
        "INSENSITIVE",
        "ISQL",
        "LC_MESSAGES",
        "LC_TYPE",
        "LEADING",
        "LENGTH",
        "LEV",
        "LOGFILE",
        "LOG_BUFFER_SIZE",
        "LOG_BUF_SIZE",
        "LONG",
        "LOWER",
        "MANUAL",
        "MAXIMUM",
        "MAXIMUM_SEGMENT",
        "MAX_SEGMENT",
        "MERGE",
        "MESSAGE",
        "MINIMUM",
        "MODULE_NAME",
        "NOAUTO",
        "NUM_LOG_BUFS",
        "NUM_LOG_BUFFERS",
        "OCTET_LENGTH",
        "OPEN",
        "OUTPUT_TYPE",
        "OVERFLOW",
        "PAGE",
        "PAGELENGTH",
        "PAGES",
        "PAGE_SIZE",
        "PARAMETER",
        #    'PASSWORD',
        "PLAN",
        "POST_EVENT",
        "QUIT",
        "RAW_PARTITIONS",
        "RDB$DB_KEY",
        "RECORD_VERSION",
        "RECREATE",
        "RECURSIVE",
        "RELEASE",
        "RESERV",
        "RESERVING",
        "RETAIN",
        "RETURN",
        "RETURNING_VALUES",
        "RETURNS",
        #    'ROLE',
        "ROW_COUNT",
        "ROWS",
        "RUNTIME",
        "SAVEPOINT",
        "SECOND",
        "SENSITIVE",
        "SHADOW",
        "SHARED",
        "SHELL",
        "SHOW",
        "SINGULAR",
        "SNAPSHOT",
        "SORT",
        "STABILITY",
        "START",
        "STARTING",
        "STARTS",
        "STATEMENT",
        "STATIC",
        "STATISTICS",
        "SUB_TYPE",
        "SUSPEND",
        "TERMINATOR",
        "TRAILING",
        "TRIGGER",
        "TRIM",
        "TRUE",
        "TYPE",
        "UNCOMMITTED",
        "UNKNOWN",
        "USING",
        "VARIABLE",
        "VERSION",
        "WAIT",
        "WEEKDAY",
        "WHILE",
        "YEARDAY",
    )
)
FIREBIRD_NONRESERVED = set(
    (
        "BACKUP",
        "BLOCK",
        "COALESCE",
        "COLLATION",
        "COMMENT",
        "DELETING",
        "DIFFERENCE",
        "IIF",
        "INSERTING",
        "LAST",
        "LEAVE",
        "LOCK",
        "NEXT",
        "NULLIF",
        "NULLS",
        "RESTART",
        "RETURNING",
        "SCALAR_ARRAY",
        "SEQUENCE",
        "STATEMENT",
        "UPDATING",
        "ABS",
        "ACCENT",
        "ACOS",
        "ALWAYS",
        "ASCII_CHAR",
        "ASCII_VAL",
        "ASIN",
        "ATAN",
        "ATAN2",
        "BACKUP",
        "BIN_AND",
        "BIN_OR",
        "BIN_SHL",
        "BIN_SHR",
        "BIN_XOR",
        "BLOCK",
        "CEIL",
        "CEILING",
        "COLLATION",
        "COMMENT",
        "COS",
        "COSH",
        "COT",
        "DATEADD",
        "DATEDIFF",
        "DECODE",
        "DIFFERENCE",
        "EXP",
        "FLOOR",
        "GEN_UUID",
        "GENERATED",
        "HASH",
        "IIF",
        "LIST",
        "LN",
        "LOG",
        "LOG10",
        "LPAD",
        "MATCHED",
        "MATCHING",
        "MAXVALUE",
        "MILLISECOND",
        "MINVALUE",
        "MOD",
        "NEXT",
        "OVERLAY",
        "PAD",
        "PI",
        "PLACING",
        "POWER",
        "PRESERVE",
        "RAND",
        "REPLACE",
        "RESTART",
        "RETURNING",
        "REVERSE",
        "ROUND",
        "RPAD",
        "SCALAR_ARRAY",
        "SEQUENCE",
        "SIGN",
        "SIN",
        "SINH",
        "SPACE",
        "SQRT",
        "TAN",
        "TANH",
        "TEMPORARY",
        "TRUNC",
        "WEEK",
    )
)

# Thanks Jonathan Lundell
MYSQL = set(
    (
        "ACCESSIBLE",
        "ADD",
        "ALL",
        "ALTER",
        "ANALYZE",
        "AND",
        "AS",
        "ASC",
        "ASENSITIVE",
        "BEFORE",
        "BETWEEN",
        "BIGINT",
        "BINARY",
        "BLOB",
        "BOTH",
        "BY",
        "CALL",
        "CASCADE",
        "CASE",
        "CHANGE",
        "CHAR",
        "CHARACTER",
        "CHECK",
        "COLLATE",
        "COLUMN",
        "CONDITION",
        "CONSTRAINT",
        "CONTINUE",
        "CONVERT",
        "CREATE",
        "CROSS",
        "CURRENT_DATE",
        "CURRENT_TIME",
        "CURRENT_TIMESTAMP",
        "CURRENT_USER",
        "CURSOR",
        "DATABASE",
        "DATABASES",
        "DAY_HOUR",
        "DAY_MICROSECOND",
        "DAY_MINUTE",
        "DAY_SECOND",
        "DEC",
        "DECIMAL",
        "DECLARE",
        "DEFAULT",
        "DELAYED",
        "DELETE",
        "DESC",
        "DESCRIBE",
        "DETERMINISTIC",
        "DISTINCT",
        "DISTINCTROW",
        "DIV",
        "DOUBLE",
        "DROP",
        "DUAL",
        "EACH",
        "ELSE",
        "ELSEIF",
        "ENCLOSED",
        "ESCAPED",
        "EXISTS",
        "EXIT",
        "EXPLAIN",
        "FALSE",
        "FETCH",
        "FLOAT",
        "FLOAT4",
        "FLOAT8",
        "FOR",
        "FORCE",
        "FOREIGN",
        "FROM",
        "FULLTEXT",
        "GRANT",
        "GROUP",
        "HAVING",
        "HIGH_PRIORITY",
        "HOUR_MICROSECOND",
        "HOUR_MINUTE",
        "HOUR_SECOND",
        "IF",
        "IGNORE",
        "IGNORE_SERVER_IDS",
        "IGNORE_SERVER_IDS",
        "IN",
        "INDEX",
        "INFILE",
        "INNER",
        "INOUT",
        "INSENSITIVE",
        "INSERT",
        "INT",
        "INT1",
        "INT2",
        "INT3",
        "INT4",
        "INT8",
        "INTEGER",
        "INTERVAL",
        "INTO",
        "IS",
        "ITERATE",
        "JOIN",
        "KEY",
        "KEYS",
        "KILL",
        "LEADING",
        "LEAVE",
        "LEFT",
        "LIKE",
        "LIMIT",
        "LINEAR",
        "LINES",
        "LOAD",
        "LOCALTIME",
        "LOCALTIMESTAMP",
        "LOCK",
        "LONG",
        "LONGBLOB",
        "LONGTEXT",
        "LOOP",
        "LOW_PRIORITY",
        "MASTER_HEARTBEAT_PERIOD",
        "MASTER_HEARTBEAT_PERIOD",
        "MASTER_SSL_VERIFY_SERVER_CERT",
        "MATCH",
        "MAXVALUE",
        "MAXVALUE",
        "MEDIUMBLOB",
        "MEDIUMINT",
        "MEDIUMTEXT",
        "MIDDLEINT",
        "MINUTE_MICROSECOND",
        "MINUTE_SECOND",
        "MOD",
        "MODIFIES",
        "NATURAL",
        "NO_WRITE_TO_BINLOG",
        "NOT",
        "NULL",
        "NUMERIC",
        "ON",
        "OPTIMIZE",
        "OPTION",
        "OPTIONALLY",
        "OR",
        "ORDER",
        "OUT",
        "OUTER",
        "OUTFILE",
        "PRECISION",
        "PRIMARY",
        "PROCEDURE",
        "PURGE",
        "RANGE",
        "READ",
        "READ_WRITE",
        "READS",
        "REAL",
        "REFERENCES",
        "REGEXP",
        "RELEASE",
        "RENAME",
        "REPEAT",
        "REPLACE",
        "REQUIRE",
        "RESIGNAL",
        "RESIGNAL",
        "RESTRICT",
        "RETURN",
        "REVOKE",
        "RIGHT",
        "RLIKE",
        "SCHEMA",
        "SCHEMAS",
        "SECOND_MICROSECOND",
        "SELECT",
        "SENSITIVE",
        "SEPARATOR",
        "SET",
        "SHOW",
        "SIGNAL",
        "SIGNAL",
        "SMALLINT",
        "SPATIAL",
        "SPECIFIC",
        "SQL",
        "SQL_BIG_RESULT",
        "SQL_CALC_FOUND_ROWS",
        "SQL_SMALL_RESULT",
        "SQLEXCEPTION",
        "SQLSTATE",
        "SQLWARNING",
        "SSL",
        "STARTING",
        "STRAIGHT_JOIN",
        "TABLE",
        "TERMINATED",
        "THEN",
        "TINYBLOB",
        "TINYINT",
        "TINYTEXT",
        "TO",
        "TRAILING",
        "TRIGGER",
        "TRUE",
        "UNDO",
        "UNION",
        "UNIQUE",
        "UNLOCK",
        "UNSIGNED",
        "UPDATE",
        "USAGE",
        "USE",
        "USING",
        "UTC_DATE",
        "UTC_TIME",
        "UTC_TIMESTAMP",
        "VALUES",
        "VARBINARY",
        "VARCHAR",
        "VARCHARACTER",
        "VARYING",
        "WHEN",
        "WHERE",
        "WHILE",
        "WITH",
        "WRITE",
        "XOR",
        "YEAR_MONTH",
        "ZEROFILL",
    )
)

MSSQL = set(
    (
        "ADD",
        "ALL",
        "ALTER",
        "AND",
        "ANY",
        "AS",
        "ASC",
        "AUTHORIZATION",
        "BACKUP",
        "BEGIN",
        "BETWEEN",
        "BREAK",
        "BROWSE",
        "BULK",
        "BY",
        "CASCADE",
        "CASE",
        "CHECK",
        "CHECKPOINT",
        "CLOSE",
        "CLUSTERED",
        "COALESCE",
        "COLLATE",
        "COLUMN",
        "COMMIT",
        "COMPUTE",
        "CONSTRAINT",
        "CONTAINS",
        "CONTAINSTABLE",
        "CONTINUE",
        "CONVERT",
        "CREATE",
        "CROSS",
        "CURRENT",
        "CURRENT_DATE",
        "CURRENT_TIME",
        "CURRENT_TIMESTAMP",
        "CURRENT_USER",
        "CURSOR",
        "DATABASE",
        "DBCC",
        "DEALLOCATE",
        "DECLARE",
        "DEFAULT",
        "DELETE",
        "DENY",
        "DESC",
        "DISK",
        "DISTINCT",
        "DISTRIBUTED",
        "DOUBLE",
        "DROP",
        "DUMMY",
        "DUMP",
        "ELSE",
        "END",
        "ERRLVL",
        "ESCAPE",
        "EXCEPT",
        "EXEC",
        "EXECUTE",
        "EXISTS",
        "EXIT",
        "FETCH",
        "FILE",
        "FILLFACTOR",
        "FOR",
        "FOREIGN",
        "FREETEXT",
        "FREETEXTTABLE",
        "FROM",
        "FULL",
        "FUNCTION",
        "GOTO",
        "GRANT",
        "GROUP",
        "HAVING",
        "HOLDLOCK",
        "IDENTITY",
        "IDENTITY_INSERT",
        "IDENTITYCOL",
        "IF",
        "IN",
        "INDEX",
        "INNER",
        "INSERT",
        "INTERSECT",
        "INTO",
        "IS",
        "JOIN",
        "KEY",
        "KILL",
        "LEFT",
        "LIKE",
        "LINENO",
        "LOAD",
        "NATIONAL ",
        "NOCHECK",
        "NONCLUSTERED",
        "NOT",
        "NULL",
        "NULLIF",
        "OF",
        "OFF",
        "OFFSETS",
        "ON",
        "OPEN",
        "OPENDATASOURCE",
        "OPENQUERY",
        "OPENROWSET",
        "OPENXML",
        "OPTION",
        "OR",
        "ORDER",
        "OUTER",
        "OVER",
        "PERCENT",
        "PLAN",
        "PRECISION",
        "PRIMARY",
        "PRINT",
        "PROC",
        "PROCEDURE",
        "PUBLIC",
        "RAISERROR",
        "READ",
        "READTEXT",
        "RECONFIGURE",
        "REFERENCES",
        "REPLICATION",
        "RESTORE",
        "RESTRICT",
        "RETURN",
        "REVOKE",
        "RIGHT",
        "ROLLBACK",
        "ROWCOUNT",
        "ROWGUIDCOL",
        "RULE",
        "SAVE",
        "SCHEMA",
        "SELECT",
        "SESSION_USER",
        "SET",
        "SETUSER",
        "SHUTDOWN",
        "SOME",
        "STATISTICS",
        "SYSTEM_USER",
        "TABLE",
        "TEXTSIZE",
        "THEN",
        "TO",
        "TOP",
        "TRAN",
        "TRANSACTION",
        "TRIGGER",
        "TRUNCATE",
        "TSEQUAL",
        "UNION",
        "UNIQUE",
        "UPDATE",
        "UPDATETEXT",
        "USE",
        "USER",
        "VALUES",
        "VARYING",
        "VIEW",
        "WAITFOR",
        "WHEN",
        "WHERE",
        "WHILE",
        "WITH",
        "WRITETEXT",
    )
)

ORACLE = set(
    (
        "ACCESS",
        "ADD",
        "ALL",
        "ALTER",
        "AND",
        "ANY",
        "AS",
        "ASC",
        "AUDIT",
        "BETWEEN",
        "BY",
        "CHAR",
        "CHECK",
        "CLUSTER",
        "COLUMN",
        "COMMENT",
        "COMPRESS",
        "CONNECT",
        "CREATE",
        "CURRENT",
        "DATE",
        "DECIMAL",
        "DEFAULT",
        "DELETE",
        "DESC",
        "DISTINCT",
        "DROP",
        "ELSE",
        "EXCLUSIVE",
        "EXISTS",
        "FILE",
        "FLOAT",
        "FOR",
        "FROM",
        "GRANT",
        "GROUP",
        "HAVING",
        "IDENTIFIED",
        "IMMEDIATE",
        "IN",
        "INCREMENT",
        "INDEX",
        "INITIAL",
        "INSERT",
        "INTEGER",
        "INTERSECT",
        "INTO",
        "IS",
        "LEVEL",
        "LIKE",
        "LOCK",
        "LONG",
        "MAXEXTENTS",
        "MINUS",
        "MLSLABEL",
        "MODE",
        "MODIFY",
        "NOAUDIT",
        "NOCOMPRESS",
        "NOT",
        "NOWAIT",
        "NULL",
        "NUMBER",
        "OF",
        "OFFLINE",
        "ON",
        "ONLINE",
        "OPTION",
        "OR",
        "ORDER",
        "PCTFREE",
        "PRIOR",
        "PRIVILEGES",
        "PUBLIC",
        "RAW",
        "RENAME",
        "RESOURCE",
        "REVOKE",
        "ROW",
        "ROWID",
        "ROWNUM",
        "ROWS",
        "SELECT",
        "SESSION",
        "SET",
        "SHARE",
        "SIZE",
        "SMALLINT",
        "START",
        "SUCCESSFUL",
        "SYNONYM",
        "SYSDATE",
        "TABLE",
        "THEN",
        "TO",
        "TRIGGER",
        "UID",
        "UNION",
        "UNIQUE",
        "UPDATE",
        "USER",
        "VALIDATE",
        "VALUES",
        "VARCHAR",
        "VARCHAR2",
        "VIEW",
        "WHENEVER",
        "WHERE",
        "WITH",
    )
)

SQLITE = set(
    (
        "ABORT",
        "ACTION",
        "ADD",
        "AFTER",
        "ALL",
        "ALTER",
        "ANALYZE",
        "AND",
        "AS",
        "ASC",
        "ATTACH",
        "AUTOINCREMENT",
        "BEFORE",
        "BEGIN",
        "BETWEEN",
        "BY",
        "CASCADE",
        "CASE",
        "CAST",
        "CHECK",
        "COLLATE",
        "COLUMN",
        "COMMIT",
        "CONFLICT",
        "CONSTRAINT",
        "CREATE",
        "CROSS",
        "CURRENT_DATE",
        "CURRENT_TIME",
        "CURRENT_TIMESTAMP",
        "DATABASE",
        "DEFAULT",
        "DEFERRABLE",
        "DEFERRED",
        "DELETE",
        "DESC",
        "DETACH",
        "DISTINCT",
        "DROP",
        "EACH",
        "ELSE",
        "END",
        "ESCAPE",
        "EXCEPT",
        "EXCLUSIVE",
        "EXISTS",
        "EXPLAIN",
        "FAIL",
        "FOR",
        "FOREIGN",
        "FROM",
        "FULL",
        "GLOB",
        "GROUP",
        "HAVING",
        "IF",
        "IGNORE",
        "IMMEDIATE",
        "IN",
        "INDEX",
        "INDEXED",
        "INITIALLY",
        "INNER",
        "INSERT",
        "INSTEAD",
        "INTERSECT",
        "INTO",
        "IS",
        "ISNULL",
        "JOIN",
        "KEY",
        "LEFT",
        "LIKE",
        "LIMIT",
        "MATCH",
        "NATURAL",
        "NO",
        "NOT",
        "NOTNULL",
        "NULL",
        "OF",
        "OFFSET",
        "ON",
        "OR",
        "ORDER",
        "OUTER",
        "PLAN",
        "PRAGMA",
        "PRIMARY",
        "QUERY",
        "RAISE",
        "REFERENCES",
        "REGEXP",
        "REINDEX",
        "RELEASE",
        "RENAME",
        "REPLACE",
        "RESTRICT",
        "RIGHT",
        "ROLLBACK",
        "ROW",
        "SAVEPOINT",
        "SELECT",
        "SET",
        "TABLE",
        "TEMP",
        "TEMPORARY",
        "THEN",
        "TO",
        "TRANSACTION",
        "TRIGGER",
        "UNION",
        "UNIQUE",
        "UPDATE",
        "USING",
        "VACUUM",
        "VALUES",
        "VIEW",
        "VIRTUAL",
        "WHEN",
        "WHERE",
    )
)


MONGODB_NONRESERVED = set(("SAFE",))

# remove from here when you add a list.
JDBCSQLITE = SQLITE
DB2 = INFORMIX = INGRES = JDBCPOSTGRESQL = COMMON

ADAPTERS = {
    "sqlite": SQLITE,
    "mysql": MYSQL,
    "postgres": POSTGRESQL,
    "postgres_nonreserved": POSTGRESQL_NONRESERVED,
    "oracle": ORACLE,
    "mssql": MSSQL,
    "mssql2": MSSQL,
    "db2": DB2,
    "informix": INFORMIX,
    "firebird": FIREBIRD,
    "firebird_embedded": FIREBIRD,
    "firebird_nonreserved": FIREBIRD_NONRESERVED,
    "ingres": INGRES,
    "ingresu": INGRES,
    "jdbc:sqlite": JDBCSQLITE,
    "jdbc:postgres": JDBCPOSTGRESQL,
    "common": COMMON,
    "mongodb_nonreserved": MONGODB_NONRESERVED,
}

ADAPTERS["all"] = reduce(lambda a, b: a.union(b), (x for x in ADAPTERS.values()))



import re

from .._compat import long
from . import adapters
from .base import SQLAdapter


@adapters.register_for("firebird")
class FireBird(SQLAdapter):
    dbengine = "firebird"
    drivers = ("kinterbasdb", "firebirdsql", "fdb", "pyodbc")

    support_distributed_transaction = True
    commit_on_alter_table = True

    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<host>[^:/]+|\[[^\]]+\])(:(?P<port>\d+))?"
        "/(?P<db>[^?]+)"
        r"(\?set_encoding=(?P<charset>\w+))?$"
    )

    def _initialize_(self):
        super(FireBird, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        user = self.credential_decoder(m.group("user"))
        password = self.credential_decoder(m.group("password"))
        if password is None:
            password = ""
        host = m.group("host")
        db = self.credential_decoder(m.group("db"))
        port = int(m.group("port") or 3050)
        charset = m.group("charset") or "UTF8"
        self.driver_args.update(
            dsn="%s/%s:%s" % (host, port, db),
            user=user,
            password=password,
            charset=charset,
        )

    def connector(self):
        return self.driver.connect(**self.driver_args)

    def test_connection(self):
        self.execute("select current_timestamp from RDB$DATABASE")

    def lastrowid(self, table):
        sequence_name = table._sequence_name
        self.execute("SELECT gen_id(%s, 0) FROM rdb$database" % sequence_name)
        return int(self.cursor.fetchone()[0])

    def create_sequence_and_triggers(self, query, table, **args):
        tablename = table._rname
        sequence_name = table._sequence_name
        trigger_name = table._trigger_name
        self.execute(query)
        self.execute("create generator %s;" % sequence_name)
        self.execute("set generator %s to 0;" % sequence_name)
        qid = self.dialect.quote_template % "id"
        self.execute(
            "create trigger %s for %s active before insert position 0 as\n"
            "begin\nif(new.%s is null) then new.%s = gen_id(%s, 1);\n"
            "end;" % (trigger_name, tablename, qid, qid, sequence_name)
        )


@adapters.register_for("firebird_embedded")
class FireBirdEmbedded(FireBird):
    REGEX_URI = (
        "^(?P<user>[^:@]+)(:(?P<password>[^@]*))?"
        r"@(?P<path>[^?]+)(\?set_encoding=(?P<charset>\w+))?$"
    )

    def _initialize_(self):
        super(FireBird, self)._initialize_()
        ruri = self.uri.split("://", 1)[1]
        m = re.match(self.REGEX_URI, ruri)
        if not m:
            raise SyntaxError("Invalid URI string in DAL")
        user = self.credential_decoder(m.group("user"))
        password = self.credential_decoder(m.group("password"))
        if password is None:
            password = ""
        pathdb = m.group("path")
        charset = m.group("charset") or "UTF8"
        self.driver_args.update(
            host="", database=pathdb, user=user, password=password, charset=charset
        )
